#########################
# Area::Structure::Room #
#########################

package Roguelike::Area::Structure::Room;

use strict;
use warnings;
use Roguelike::Game;
use Roguelike::Utility;

use Roguelike::Tilemap;
use Roguelike::Area;

use Exporter ();
our @ISA = ("Exporter");
our @EXPORT = do {
  no strict "refs";
  grep defined &$_, keys %{ __PACKAGE__ . "::" };
};

sub new {
  shift;  # Who cares?
  my $height = shift;
  my $width = shift;
  # Deal with the offset bug LATER.
  my $map = new Roguelike::Tilemap $height, $width, ".";
  my ($ncut, $scut, $wcut, $ecut) = (0, 0, 0, 0);
  foreach my $corner (1 .. 4) {
    if (percent 50) {
      my $vcut = random($map->height / 2);
      my $hcut = random($map->width / 2);
      foreach my $y (0 .. $vcut - 1) { 
        foreach my $x (0 .. $hcut - 1) {
          $map->{Map}[$y][$x]{_} = " " if $corner == 1;
          $map->{Map}[$y][-$x - 1]{_} = " " if $corner == 2;
          $map->{Map}[-$y - 1][$x]{_} = " " if $corner == 3;
          $map->{Map}[-$y - 1][-$x - 1]{_} = " " if $corner == 4;
        }
      }
      if ($corner == 1 or $corner == 2) {
        $ncut = $vcut if $vcut > $ncut;
      } elsif ($corner == 3 or $corner == 4) {
        $scut = $vcut if $vcut > $scut;
      }
      if ($corner == 1 or $corner == 3) {
        $wcut = $hcut if $hcut > $wcut;
      } elsif ($corner == 2 or $corner == 4) {
        $ecut = $hcut if $hcut > $ecut;
      }
    }
  }
  $map->border;
  $map->{Exits} = findwalls($map, $ncut, $scut, $wcut, $ecut);
  $map->{Exits}{Type} = "Room";
  # Hawt spots
  $map->{Exits}{Hotspots} = [];
  {
    my $dat = chooserand(@{ $map->{Exits}{N} });
    my $y = $dat->{Y1} + random(1, 4);  # TWEAKYWEAKYWOOOOOOOOOOOOO
    my $x = random($dat->{X1} + 1, $dat->{X2} - 1);
    next if $y > $map->height or $y < 0;
    next if $x > $map->width or $x < 0;
    push @{ $map->{Exits}{Hotspots} }, [$y, $x] if $map->{Map}[$y][$x]{_} eq ".";
  }
  {
    my $dat = chooserand(@{ $map->{Exits}{W} });
    my $y = random($dat->{Y1} + 1, $dat->{Y2} - 1);
    my $x = $dat->{X1} + random(1, 4);  # TWEAKYWEAKYWOOOOOOOOOOOOO
    next if $y > $map->height or $y < 0;
    next if $x > $map->width or $x < 0;
    next unless defined $map->{Map}[$y][$x];
    push @{ $map->{Exits}{Hotspots} }, [$y, $x] if $map->{Map}[$y][$x]{_} eq ".";
  }
  return $map;
}

42;
