##############################
# StdLib::Character::Monster #
##############################

package Roguelike::StdLib::Character::Monster;

use strict;
use warnings;
use Roguelike::Game;
use Roguelike::Utility;

use Roguelike::Container;

# Definition

my $monster = $Game->{Templates}{Character}->new(
  Symbol => "@",  # Ah, the nostalgia...
  Type => "Monster"
);

# Reactions

# None yet. :(

# Actions

sub UTIL_Init {
  my $self = shift;
  # Whoa, wait, are we an instance? If we are, we have Z, Y, and X. No base.
  if (defined $self->{Z} and defined $self->{Y} and defined $self->{X}) {
    # Set up some stuff up...
    $self->{Inv} = new Roguelike::Container 0, $self;
    if ($self->go( $self->g("Z"), $self->g("Y"), $self->g("X") ) == -1) {
      #debug "$self->{Z}, $self->{Y}, $self->{X}... why can't we go?";
      #debug $Game->{Levels}[$self->{Z}]{Map}[$self->{Y}][$self->{X}];
      return -1;
    }
    # Register ourselves in the area's list.
    unless ($self->g("Area")) {
      #debug "$self->{Z}, $self->{Y}, $self->{X}... WHERE ARE WE?";
      #debug $Game->{Levels}[$self->{Z}]{Map}[$self->{Y}][$self->{X}];
      return -1;
    } else {
      push @{ $self->g("Area")->{Monsters} }, $self;
      # And queue ourselves
      $self->schedule(1, sub { my $self = shift; $self->Input; });
    }
  }
}

sub PRE_die { (0, "die", "-") }

sub die {
  my $self = shift;
  my $selfdestruct = shift;
  # Erase ourselves from the map.
  $self->tile->{Char} = undef;
  render($self->tile);
  # Remove self from list.
  # Just loop through for now. Whatever.
  foreach (0 .. $#{ $self->g("Area.Monsters") }) {
    splice(@{ $self->g("Area.Monsters") }, $_, 1) if $self->g("ID") ==
    $self->g("Area.Monsters.$_.ID");
  }
  # Now do the same for the queue.
  foreach (0 .. $#{ $Game->{Queue} }) {
    my $tmp = [];
    foreach (0 .. $#{ $Game->{Queue} }) {
      push @$tmp, $Game->{Queue}[$_] unless $Game->{Queue}[$_]{Actor}->g("ID") == $self->g("ID");
    }
    @{ $Game->{Queue} } = @{ $tmp };
  }
  # For now, only the player kills monsters... This is so messy!
  if ($selfdestruct) {
    $self->msg("<Red>[subj] is obliterated.");
  } else {
    $Player->msg("<Red>[subj] kill the [1]!", $self);
    # Experience.
    $Player->gainexp( $self->g("Exp") );
  }
  $Game->{Moved} = 1;
  undef $self;  # Weak references better be good for something. Assume there
                # are no others.
}

sub UTIL_Input {
  my $self = shift;
  my ($energy, $action, @args);
  # We active?
  if ($Game->{Active}{ $self->{ID} }) {
    my @distance = $self->distance($Player);
    # Here we go!
    HANDLE: foreach my $handle ($self->a("Behavior")) {
      # Test the conditions.
      CHECK: foreach my $state (@{ $handle->{Condition} }) {
        my ($cmd, @args) = @$state;
        if ($cmd eq "TP_lower") {
          next HANDLE unless $self->{TP} < $args[0];
        } elsif ($cmd eq "adjacent") {
          next HANDLE unless $distance[0] < 2 and $distance[1] < 2;
        } elsif ($cmd eq "near") {
          next HANDLE unless $distance[0] + $distance[1] <= $args[0];
        } elsif ($cmd eq "else") {
          next; # Success, always.
        } elsif ($cmd eq "TP_higher") {
          next HANDLE unless $self->{TP} > $args[0];
        } elsif ($cmd eq "percent") {
          next HANDLE unless percent($args[0]);
        } elsif ($cmd eq "lined_up") {
          foreach my $dir (qw(h j k l y u b n)) {
            if (my $target = ($self->{Area}->trace($self->{Y}, $self->{X}, $dir))[0]) {
              next CHECK if $target->is("Player");
            }
          }
          next HANDLE;
        } elsif ($cmd eq "health_low") {
          debug "hp is $self->{HP}, arg is $args[0]";
          next unless $self->{HP} > $args[0];
        } else {
          debug "Er, there is no AI handler for the $cmd condition!";
        }
      }
      my ($cmd, @args) = @{ $handle->{Action} };
      if ($cmd eq "run") {
        $self->{Path} = []; # no jumping!
        my $py = $Player->{Y};
        my $px = $Player->{X};
        # The crappiness of this ASTOUNDS me.
        my $far;
        my ($Y, $X);
        foreach my $dir (qw(h j k l y u b n)) {
          my ($y, $x) = keycoords($dir, $self->{Y}, $self->{X});
          next if $self->g("Area.Map.$y.$x._") ne "+" and $self->g("Area")->barrier($y, $x);
          next if $self->g("Area.Map.$y.$x.Char");
          my $ourdistance = abs($y - $py) + abs($x - $px);
          if ((!defined $far) or $ourdistance > $far) {
            ($Y, $X) = ($y, $x);
            $far = $ourdistance;
          }
        }
        unless (defined $Y and defined $X) {
          $self->queue("idle")->();
        } else {
          my @results = $self->_go($Y, $X);
          $self->queue("_$results[1]_", @results[2 .. 4])->();
        }
      } elsif ($cmd eq "attack") {
        $self->queue("attack", @args, $Player)->();
      } elsif ($cmd eq "stupid_chase") {
        my $py = $Player->{Y};
        my $px = $Player->{X};
        # The crappiness of this ASTOUNDS me.
        my $far;
        my ($Y, $X);
        foreach my $dir (qw(h j k l y u b n)) {
          my ($y, $x) = keycoords($dir, $self->{Y}, $self->{X});
          next if $self->g("Area.Map.$y.$x._") ne "+" and $self->g("Area")->barrier($y, $x);
          next if $self->g("Area.Map.$y.$x.Char");
          my $ourdistance = abs($y - $py) + abs($x - $px);
          if ((!defined $far) or $ourdistance < $far) {
            ($Y, $X) = ($y, $x);
            $far = $ourdistance;
          }
        }
        unless (defined $Y and defined $X) {
          $self->queue("idle")->();
        } else {
          my @results = $self->_go($Y, $X);
          $self->queue("_$results[1]_", @results[2 .. 4])->();
        }
      } elsif ($cmd eq "smart_chase") {
        # OK, move. Do we need to pathfind?
        unless ($self->g("Path.0")) {
          my @path = $self->g("Area")->pathfind($self->{Y}, $self->{X}, $Player->{Y}, $Player->{X});
          if ($path[0] == -1) {
            $self->queue("idle")->();
            last;
          }
          # Save the list, PLEASE! We don't all have 4GHz boxes.
          $self->{Path} = \@path;
        }
        # So now we move.
        # But wait. Is there a non-player character in our way? We could be
        # patient.
        my ($y, $x) = @{ shift @{ $self->g("Path") } };
        if ($self->g("Area.Map.$y.$x.Char") and
            $self->g("Area.Map.$y.$x.Char.ID") != $Player->{ID}
           )
        {
          $self->queue("_idle_")->();
          # So try again!
          unshift @{ $self->{Path} }, [$y, $x];
          return 1;
        }
        my $repeat = $self->g("Area.Map.$y.$x._") eq "+" ? 1 : 0;
        my @results = $self->_go($y, $x);
        $self->queue("_$results[1]_", @results[2 .. 4])->();
        unshift @{ $self->{Path} }, [$y, $x] if $repeat;
      } else {
        debug "Er, there is no AI handler for the $cmd action!";
      }
      # Best place for this.
      @distance = $self->distance($Player);
      $self->{Path} = [] if $distance[0] < 2 and $distance[1] < 2;
      last;
    }
    return 1;
  } else {
    $self->queue("idle")->();
    return 1;
  }
}

define($monster, ["die"]);

42;
