#############################
# StdLib::Character::Player #
#############################

package Roguelike::StdLib::Character::Player;

use strict;
use warnings;
use Roguelike::Game;
use Roguelike::Utility;

use Roguelike::Tilemap;
use Roguelike::Container;
use Storable;

# Definition

my $pc = $Game->{Templates}{Character}->new(
  Name => "Player",
  Symbol => "@",  # Ah, the nostalgia...
  #Unique => 1,    # How's this tie in with races and classes?
  Actions => {
    exit => sub {
      # This subroutine name might not work so well with the exporter. :P
      # Temporary, this is, of course
      exit;
    }
  },
  Level => 0,
  Exp => 0,
  ExpNeeded => 25,
  Techniques => {},
  ModeLaugh => 0,  # EWWWWWW
  Type => "PlayerChar"
);

# Reactions

$pc->react(
  to => ["After", "go"],
  by => sub {
    my $self = shift;
    $self->list(1);
    return 1;
  },
  Name => "LSTILE"
);

# Actions

sub UTIL_Init {
  my $self = shift;
  # We're always an instance, right?
  # Wrong.
  if ($self->{PCInstance}) {
    # Set up some stuff up...
    $self->{Inv} = new Roguelike::Container 1, $self;
    $self->schedule(1, sub { $self->Input(@_) });
    unless (defined $self->{Y} and defined $self->{X}) {  # No base
      my $dat = chooserand(@{ $Game->{Levels}[ $self->g("Z") ]{StairsUp} });
      ($self->{Y}, $self->{X}) = @$dat;
    }
    $self->go( $self->g("Z"), $self->g("Y"), $self->g("X") );
    $self->levelup;
    $self->{Turns} = -1;
  }
  return 1;
}

sub PRE_exit { 
  shift;
  my $force = shift;
  my $in;
  unless ($force) {
    say "Do you really want to quit? (Yn)";
    $Game->{UI}->msgrefresh;
    $in = $Game->{UI}->prompt;
  }
  if ($force or $in eq "Y") {
    return (0, "exit");
  } else {
    return -1;
  }
}

sub UTIL_Input {
  my $self = shift;
  my ($dothis, $energy);
  $Player->{Turns}++;
  $Game->{UI}->refresh;
  $Game->{UI}->statrefresh;
  while (1) {
    $energy = 1;
    if ($Game->{Suppress}) {
      $Game->{Suppress} = 0;
    } else {
      $Game->{UI}->msgrefresh();
    }
    my $in = $Game->{UI}->prompt;
    my $action = $Game->{UI}{Keymap}{$in};
    unless ($action) {
      say "The $in key is not bound.";
      next;
    }
    $action = "_$action";
    my ($meth, @args);
    my @results = $self->$action();
    if (@results == 1) {
      $energy = -1;
    } else {
      ($energy, $meth, @args) = @results;
    }
    if ($energy == -13) { # NEW MAGIC NUMBER!!!
      $self = $args[0];
      next;
    }
    next if $energy == -1;
    $meth = "_${meth}_";
    if ($energy == 0) {
      $self->$meth(@args);
      next;
    } elsif ($energy == 1) {
      $self->queue($meth, @args)->();
      if ($Game->{Propogate}) {
        $Game->{Propogate} = 0;
        return (-16, $self);
      }
      return 1;
    } else {
      $dothis = $self->queue($meth, @args);
      last;
    }
  }
  $self->schedule($energy, $dothis);
  return 1;
}

sub PRE_test { (0, "test") }

sub test {
  #complain "Multi-colored <green>monstrosity!";
  #say "boring plain grey";
  #say "<purple>foo<white>bar";
  #say "Long, infinitely LONGISH sorts of lines that appear to go on for
  #     infinity captivate and... imaginate? the imagination for all of
  #     eternity, as visible in the eye of the <green>beholder<aqua>WEE!";
  #say "But when light fades... will the hero rise again?";
  #say "<RED>bright colors <GREEN>as opposed to <grey>dim colors";
  #$Game->save();
  $Player->levelup;
}

sub PRE_test2 { (0, "test2") }

sub test2 {
  my $self = shift;
  my $map = $self->g("Area");
  open FOO, ">demomap";
  foreach my $y (0 .. $map->height) {
    foreach my $x (0 .. $map->width) {
      print FOO $map->{Map}[$y][$x]{_};
    }
    print FOO "\n";
  }
  close FOO;
  return 1;
}

sub PRE_msglog { (0, "msglog") }

sub msglog {
  $Game->{UI}->display(0, @{ $Game->{MsgLog} });
}

sub PRE_inv { (0, "inv", "-") }

sub inv {
  my $self = shift;
  my $mode = shift;
  $mode = defined $mode ? $mode : 0;
  my @categories;
  if (@_) {
    @categories = @_;
  } else {
    @categories = keys %{ $Game->{Config}{Category} };
  }
  my @lines = ("<yellow>Inventory", "");
  foreach my $cat (@categories) {  # Dog... GRANDMA!!!
    push @lines, " <RED>$cat" if keys %{ $Player->{Inv}{Category}{$cat} };
    foreach (sort keys %{ $Player->{Inv}{Category}{$cat} }) {
      push @lines, "  $_ - " . $Player->{Inv}->get($_)->invname;
    }
  }
  return $Game->{UI}->display($mode, @lines);
}

sub PRE_list { (0, "list", "-") }

sub list {
  my $self = shift;
  my $indirect = shift;
  my $inv = $self->tile->{Inv};
  say "There's a staircase leading up here." if $self->tile->{_} eq "<";
  say "There's a staircase leading down here." if $self->tile->{_} eq ">";
  say "You stand in a doorway." if $self->tile->{_} eq "'" and !$indirect;
  if ($inv->all == 1) {
    say "You see here " . $inv->get(0)->generalname . ".";
  } elsif ($inv->all > 1) {
    say "You see here:";
    foreach (reverse $inv->all) {
      say " " . $_->generalname;
    }
  } else {
    say "There's nothing here." unless $indirect;
  }
  return 1;
}

sub PRE_viewinv {
  my $self = shift;
  my $item = $Game->{UI}->selectinv($self, "Examine what? (? to list items)");
  return -1 if $item == -1;
  $Game->{UI}->msgrefresh;
  return (0, "examine", $item);
}

sub PRE_examine { (0, "examine", "-") }

sub examine {
  my ($self, $item) = @_;
  my @lines;
  push @lines, "<Aqua>" . $item->g("Index") . " - " . $item->invname, "";
  push @lines, $item->g("Desc"), "" if $item->g("Desc");
  if ($item->g("Wearable") or $item->g("Wieldable")) {
    push @lines, "<red>Power: <grey>" . $item->g("Power"),
                 "<red>Bonus Modifier:<grey> " . $item->g("Mod"),
                 "";
  }
  return $Game->{UI}->display(1, @lines);
}

sub PRE_examinemons { (0, "examinemons", "-") }

sub examinemons {
  my ($self, $monster) = @_;
  my @lines;
  if ($monster->is("Player")) {
    push @lines, "<Aqua>Player";
  } else {
    push @lines, "<Aqua>" . $monster->name;
  }
  push @lines, "", $monster->g("Desc"), "";
  push @lines, "<red>HP: <grey>" . $monster->g("HP");
  push @lines, "<red>TP: <grey>" . $monster->g("TP");
  push @lines, "<red>Strength: <grey>" . $monster->g("Str");
  push @lines, "<red>Defence: <grey>" . $monster->g("Def");
  push @lines, "<red>Experience: <grey>" . $monster->g("Exp");
  return $Game->{UI}->display(1, @lines);
}

sub PRE_die { (0, "die") }

sub die {
  my $self = shift;
  say "You die...";
  $Game->{UI}->refresh;
  $Game->{UI}->msgrefresh;
  $Game->{UI}->statrefresh;
  $Game->{UI}->prompt;
  $Game->{UI}->msgbox(1,
    "You awaken to the same life you had when you entered the dream. You feel
     as if you should go back and try again."
  );
  $Player->exit(1);
}

sub UTIL_gainexp {
  my $self = shift;
  my $exp = shift;
  # Add it.
  $self->{Exp} += $exp;
  # Enough?
  my $carryover = 0;
  if ($self->{Exp} >= $self->g("ExpNeeded")) {
    $carryover = $self->{Exp} - $self->g("ExpNeeded");
    $self->{Exp} = 0;
    $self->levelup;
  }
  $self->{Exp} += $carryover; # We'll *assume* we don't have enough carryover
                              # to actually make us go another level!
  return 1;
}

sub UTIL_levelup {
  my $self = shift;
  $self->{Level}++;
  $self->msg("<Green>[subj] [have/has] gained a level!");
  $self->{ExpNeeded} = 25 * $self->{Level};
  # Raise mah stats
  my $hp = min(2, 5, 10);
  my $tp = random(2, 4);
  my $str = random(1, 3);
  my $def = random(1, 3);
  $self->{HP} += $hp;
  $self->{MaxHP} += $hp;
  $self->{TP} += $tp;
  $self->{MaxTP} += $tp;
  $self->{Str} += $str;
  $self->{MaxStr} += $str;
  $self->{Def} += $def;
  $self->{MaxDef} += $def;
  # Update technique proficency
  foreach (keys %{ $self->g("Techniques") }) {
    $self->{Techniques}{$_}{Success}++ unless $self->{Techniques}{$_}{Success} == 3;
  }
  signal("When", $self, "gainlevel", $self->{Level});
  return 1;
}

sub PRE_performtech {
  my $self = shift;
  # We're _totally_ UI man.
  my $tech = $Game->{UI}->selecttech;
  return -1 if $tech eq -1;
  # But can you carry it out? :P
  my $success = 0;
  $success = 33 if $Player->g("Techniques.$tech.Success") == 1;
  $success = 66 if $Player->g("Techniques.$tech.Success") == 2;
  $success = 99 if $Player->g("Techniques.$tech.Success") == 3;
  my $name = $Player->g("Techniques.$tech.Name");
  # Enough TP?
  if ($self->g("TP") < $self->g("Attacks.$name.TP")) {
    $self->msg("[subj] doesn't have enough TP to $name!");
    return -1;
  }
  unless ( percent($success) ) {
    say "You fail to perform $name!";
    # Still deduct TP.
    $self->{TP} -= $self->g("Attacks.$name.TP");
    return (-42, "idle"); # The move still counts
  }
  # DELEGATE.
  return (-42, $name);
}

sub PRE_switchweapon {
  my $self = shift;
  # Get all the weapons, sorted by letter.
  my @weapons;
  if ($self->h("Inv.Category.Weapons")) {
    @weapons = map {
      $self->g("Inv")->get($_)
    } sort keys %{ $self->g("Inv.Category.Weapons") };
  } else {
    say "You don't even have any weapons!";
    return -1;
  }
  my $new;
  if (my $weapon = $self->g("Equipment.Weapon")) {
    # The first one that isn't $weapon
    foreach (@weapons) {
      next if $_->g("ID") == $weapon->g("ID");
      $new = $_;
      last;
    }
    unless ($new) {
      say "You have no other weapons to switch to!";
      return -1;
    }
  } else {
    $new = shift @weapons;
  }
  return (-42, "wield", $new);
}

sub PRE_lookmons {
  my $self = shift;
  say "There aren't any monsters around, but go ahead." unless keys %{ $Game->{Active} };
  $Game->{UI}->target;
  return -1;
}

sub PRE_debug {
  say "Do you really want to skip to the ending? (Yn)";
  $Game->{UI}->msgrefresh;
  my $in = $Game->{UI}->prompt;
  if ($in eq "Y") {
    # This is such a messy hack.
    foreach ($Player->{Z} + 1 .. 20) {
      $Game->{Levels}[$_] = new Roguelike::Tilemap 1, 1;
      $Game->{Levels}[$_]{ID} = $_;
    }
    $Player->{Z} = 20;
    $Player->tile->{_} = ">";
    $Player->tile->{Stair} = 0;
    $Player->descend;
  }
  return -1;
}

sub PRE_help {
   $Game->{UI}->msgbox(1,
  "Movement is done with the vi/nethack keys: <Blue>hjklyubn",
  "<Blue>Control+P<black> should open up the message log.",
  "At a screen like this, only space will proceed. At prompts for items, letters   work. <Blue>?<black> will bring up a list of relevant items.",
  "<Blue>,<black> will take an item, <Blue>i<black> brings up inventory,
  <Blue>d<black> drops, <Blue>;<black> lists current items, <Blue>.<black>
  idles, <Blue>e<black> eats, <Blue>r<black> reads, <Blue>w<black> wields
  (weapons only), <Blue>W<black> wears (other equipment), <Blue>T<black>
  discards equipment, <Blue>><black> and <Blue><<black> descend or ascend,
  <Blue>v<black> lets you examine an item, <Blue>x<black> allows you to examine
  monsters, <Blue>a<black> lets you perform one of your special abilities (which  you'll learn every few levels), <Blue>f<black> fires a ranged weapon,
  and <Blue>Q<black> exits."
  );
  $Game->{UI}->msgbox(2);
  return -1;
}

sub PRE_refresh {
  $Game->{UI}{Interface}->clearjunk;
  $Player->recenter;
  return -1;
}

sub PRE_recenter {
  $Game->{ForceRefresh} = 1;
  $Game->{UI}{Rows} = [];
  $Game->{UI}{Cols} = [];
  $Game->{ForceRefresh} = 1;
  $Game->{UI}->refresh;
  return -1;
}

sub PRE_save {
  # Set up temporary directory.
  mkdir "game";
  mkdir "game/boxes";
  mkdir "game/maps";
  mkdir "game/objs";
  mkdir "game/queue";
  $Storable::Deparse = 1; # Coderefs turn up in unlikely places.
  # Serialize objects.
  # For some reason, we screw with the list!
  my @ls = @{ $Game->{IDs} };
  foreach my $obj (@ls) {
    if (!$obj->{Type} or $obj->{PCInstance}) {
      next unless ref $obj eq "Roguelike::Object";  # Weird.
      say "<Green>Saving object $obj->{ID}/$#ls";
      $Game->{UI}->msgrefresh;
      store serialize($obj), "game/objs/$obj->{ID}";
    }
  }
  # OK, now the containers...
  foreach my $box (@{ $Game->{Containers} }) {
    next unless ref $box eq "Roguelike::Container"; # WEIRDER
    # Do we even have to?! :P
    next unless @{ $box->{List} };
    say "<Green>Serializing container $box->{ID}/$#{ $Game->{Containers} }";
    $Game->{UI}->msgrefresh;
    store serialize($box), "game/boxes/$box->{ID}";
  }
  # Finally the maps.
  foreach my $map (@{ $Game->{Levels} }) {
    say "<Green>Serializing map $map->{ID}/$#{ $Game->{Levels} }";
    $Game->{UI}->msgrefresh;
    store serialize($map), "game/maps/$map->{ID}";
  }
  # Er, now the queue it seems. Geez.
  # This could be better... subrefs suck. Maybe have 'input' or hp++?
  foreach (0 .. $#{ $Game->{Queue} }) {
    my $event = $Game->{Queue}[$_];
    say "<Green>Serializing event $_/$#{ $Game->{Queue} }";
    $Game->{UI}->msgrefresh;
    store serialize($event), "game/queue/$_";
  }
  $Storable::Deparse = 0;
  # Tar it on up. Screw platform independence.
  system("tar", "czf", "game.tgz", "game") == 0 or CORE::die "Couldn't package archive! $!";
  # Nuke temp dirs.
  unlink glob "game/*/*";
  rmdir "game/boxes";
  rmdir "game/objs";
  rmdir "game/maps";
  rmdir "game/queue";
  rmdir "game";
  return -1;
}

sub PRE_load {
  # Anything even there?
  unless (-e "game.tgz") {
    rsay "There is no saved game in your current directory!";
    return -1;
  }
  # Uncompress archive.
  system("tar", "xzf", "game.tgz") == 0 or CORE::die "Couldn't unpack archive! $!";
  my $self = shift;
  # Load all serialized data...
  # Objects first.
  say "<Green>Loading objects...";
  $Game->{UI}->msgrefresh;
  foreach my $file (glob "game/objs/*") {
    my $dat = retrieve($file);
    my $base = delete $dat->{Base};
    my $obj = $dat;
    $obj->{Base} = $Game->{Content}{$base};
    # I don't think I want to create a new object from the template... No...
    # Because we've got lots of data that shouldn't be fiddled with.
    #my $obj = $Game->{Content}{$base}->new;
    ## Restore data.
    #while (my ($key, $value) = each(%$dat)) {
    #  $obj->{$key} = $value;
    #}
    $obj->{Loaded} = 1;
    $Game->{IDs}[$obj->{ID}] = $obj;
  }
  unlink glob "game/objs/*";
  # Containers...
  say "<Green>Loading containers...";
  $Game->{UI}->msgrefresh;
  foreach my $file (glob "game/boxes/*") {
    # Did we even have to save one?
    my $box = retrieve($file);
    $box->{Loaded} = 1;
    $Game->{Containers}[$box->{ID}] = $box;
  }
  unlink glob "game/boxes/*";
  # Then maps...
  say "<Green>Loading maps...";
  $Game->{UI}->msgrefresh;
  foreach my $file (glob "game/maps/*") {
    my $map = retrieve($file);
    $map->{Loaded} = 1;
    $Game->{Levels}[$map->{ID}] = $map;
  }
  unlink glob "game/maps/*";
  # The queue...
  say "<Green>Loading queue...";
  $Game->{UI}->msgrefresh;
  $Storable::Eval = 1;
  foreach my $file (glob "game/queue/*") {
    my $event = retrieve($file);
    $event->{Loaded} = 1;
    $file =~ s!^game/queue/!!;
    $Game->{Queue}[$file] = $event;
  }
  $Storable::Eval = 0;
  unlink glob "game/queue/*";
  # Nuke temp stuff.
  rmdir "game/boxes";
  rmdir "game/objs";
  rmdir "game/maps";
  rmdir "game/queue";
  rmdir "game";
  unlink "game.tgz";
  # Now walk through all these things and turn placeholders into references.
  say "<Green>Restoring references...";
  $Game->{UI}->msgrefresh;
  # Objects first.
  foreach my $obj (@{ $Game->{IDs} }) {
    next unless $obj->{Loaded};
    unserialize($obj);
    $Player = $obj if $obj->{PCInstance};
    delete $obj->{Loaded};
  }
  # Containers.
  foreach my $box (@{ $Game->{Containers} }) {
    next unless $box->{Loaded};
    unserialize($box);
  }
  # Maps.
  foreach my $map (@{ $Game->{Levels} }) {
    next unless $map->{Loaded};
    unserialize($map);
    delete $map->{Loaded};
    # Grrr... this is temporary, I suppose.
    foreach my $y (0 .. $map->height) {
      foreach my $x (0 .. $map->width) {
        if (ref $map->{Map}[$y][$x]{Inv} eq "HASH" or 
            (!$map->{Map}[$y][$x]{Inv}{Loaded})
           )
        {
          $map->{Map}[$y][$x]{Inv} = new Roguelike::Container 0, $map->{Map}[$y][$x];
        }
        # Why not?
        $map->{Map}[$y][$x]{Inv}{Owner} = $map->{Map}[$y][$x];
        delete $map->{Map}[$y][$x]{Inv}{Loaded};
      }
    }
  }
  # The Queue.
  foreach my $event (@{ $Game->{Queue} }) {
    next unless $event->{Loaded};
    unserialize($event);
    delete $event->{Loaded};
  }
  say "<Green>Loaded!";
  $Game->{UI}->msgrefresh(1);
  $Player->recenter;
  $Game->{UI}->statrefresh;
  $self = $Player;  # WEIRD
  $Game->{Propogate} = 1;
  return (-13, "headache", $self);
}

define($pc, [
  "test", "msglog", "inv", "list", "redo", "test2", "examine", "die",
  "examinemons"
]);

42;
